use std::{
    convert::Infallible,
    error::Error,
    fmt::Display,
    num::ParseIntError,
    ops::Deref,
    path::{Path, PathBuf},
    process::Command,
    slice::Iter,
    str::FromStr,
};

use crate::fs::{
    id_mapping::{IdMapping as OriginalIdMapping, IdRange, ParseIdRangeError},
    mount::MountOptions,
    AsIter, FileSystem, Mountpoint,
};
use crate::macros::*;

use super::Stack;

macro_rules! impl_deref {
    ($name:ident($inner:ty)) => {
        impl Deref for $name {
            type Target = $inner;

            fn deref(&self) -> &Self::Target {
                &self.0
            }
        }
    };
}

macro_rules! impl_fromstr {
    ($name:ident($inner:ty), $err:ty) => {
        impl FromStr for $name {
            type Err = $err;

            fn from_str(s: &str) -> Result<Self, Self::Err> {
                Ok(Self(<$inner>::from_str(s)?))
            }
        }
    };
}

macro_rules! impl_display {
    ($name:ident) => {
        impl Display for $name {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                write!(f, "{}", self.0)
            }
        }
    };
}

macro_rules! impl_wrapper_struct {
    ($name:ident($inner:ty), $err:ty) => {
        impl_deref!($name($inner));
        impl_fromstr!($name($inner), $err);
        impl_display!($name);
    };
}

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseMaxIdleThreadsError(ParseIntError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseMaxThreadsError(ParseIntError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSquashToIdError(ParseIntError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSquashToUidError(ParseSquashToIdError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseSquashToGidError(ParseSquashToIdError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseUidMappingError(ParseIdRangeError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct ParseGidMappingError(ParseIdRangeError);

impl_wrapper_err_for!(
    ParseMaxIdleThreadsError(ParseIntError),
    "invalid max idle threads"
);

impl_wrapper_err_for!(ParseMaxThreadsError(ParseIntError), "invalid max threads");
impl_wrapper_err_for!(ParseSquashToIdError(ParseIntError), "invalid id");
impl_wrapper_err_for!(ParseSquashToUidError(ParseSquashToIdError), "invalid uid");
impl_wrapper_err_for!(ParseSquashToGidError(ParseSquashToIdError), "invalid gid");
impl_wrapper_err_for!(ParseUidMappingError(ParseIdRangeError), "invalid uid");
impl_wrapper_err_for!(ParseGidMappingError(ParseIdRangeError), "invalid gid");

#[derive(Clone, Debug, Default)]
pub struct DirIter<'a> {
    inner: Iter<'a, PathBuf>,
}

impl<'a> Iterator for DirIter<'a> {
    type Item = &'a Path;

    fn next(&mut self) -> Option<Self::Item> {
        self.inner.next().map(PathBuf::as_path)
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct LowerDirs(Vec<PathBuf>);

impl<A: Into<PathBuf>> FromIterator<A> for LowerDirs {
    fn from_iter<T: IntoIterator<Item = A>>(iter: T) -> Self {
        Self(iter.into_iter().map(Into::into).collect())
    }
}

impl<'a> IntoIterator for &'a LowerDirs {
    type Item = &'a Path;
    type IntoIter = DirIter<'a>;

    fn into_iter(self) -> Self::IntoIter {
        DirIter {
            inner: self.as_slice().iter(),
        }
    }
}

impl Deref for LowerDirs {
    type Target = Vec<PathBuf>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl Display for LowerDirs {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut iter = self.0.iter();
        if let Some(head) = iter.next() {
            write!(f, "{}", head.to_string_lossy())?;

            for item in iter {
                write!(f, ",{}", item.to_string_lossy())?;
            }
        }

        Ok(())
    }
}

impl FromStr for LowerDirs {
    type Err = Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(
            s.split(',')
                .filter(|s| !s.is_empty())
                .map(Into::into)
                .collect(),
        ))
    }
}

impl From<Vec<&str>> for LowerDirs {
    fn from(value: Vec<&str>) -> Self {
        Self(value.into_iter().map(PathBuf::from).collect())
    }
}

impl From<Vec<String>> for LowerDirs {
    fn from(value: Vec<String>) -> Self {
        Self(value.into_iter().map(PathBuf::from).collect())
    }
}

impl From<Vec<&Path>> for LowerDirs {
    fn from(value: Vec<&Path>) -> Self {
        Self(value.into_iter().map(Path::to_path_buf).collect())
    }
}

impl From<Vec<PathBuf>> for LowerDirs {
    fn from(value: Vec<PathBuf>) -> Self {
        Self(value)
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, Default)]
pub struct MaxIdleThreads(Option<usize>);

impl FromStr for MaxIdleThreads {
    type Err = ParseMaxIdleThreadsError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "-1" => Ok(Self(None)),
            s => Ok(s.parse::<usize>().map(Some).map(Self)?),
        }
    }
}

impl Display for MaxIdleThreads {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.0 {
            Some(n) => write!(f, "{n}"),
            None => write!(f, "-1"),
        }
    }
}

impl_deref!(MaxIdleThreads(Option<usize>));

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct MaxThreads(usize);

impl Default for MaxThreads {
    fn default() -> Self {
        Self(10)
    }
}

impl FromStr for MaxThreads {
    type Err = ParseMaxThreadsError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(usize::from_str_radix(s, 10).map(Self)?)
    }
}

impl_deref!(MaxThreads(usize));
impl_display!(MaxThreads);

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct SquashToId(usize);

impl_deref!(SquashToId(usize));
impl_display!(SquashToId);

impl FromStr for SquashToId {
    type Err = ParseSquashToIdError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(s.parse::<usize>()?))
    }
}

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct SquashToUid(SquashToId);

impl_wrapper_struct!(SquashToUid(SquashToId), ParseSquashToUidError);

#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct SquashToGid(SquashToId);

impl_wrapper_struct!(SquashToGid(SquashToId), ParseSquashToGidError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct IdMapping(OriginalIdMapping);

impl IdMapping {
    pub fn new(ranges: impl IntoIterator<Item = impl Into<IdRange>>) -> Self {
        Self(OriginalIdMapping::new(ranges.into_iter(), ','))
    }
}

impl From<OriginalIdMapping> for IdMapping {
    fn from(value: OriginalIdMapping) -> Self {
        Self(value)
    }
}

impl_wrapper_struct!(IdMapping(OriginalIdMapping), ParseIdRangeError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct UidMapping(IdMapping);

impl_wrapper_struct!(UidMapping(IdMapping), ParseUidMappingError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub struct GidMapping(IdMapping);

impl_wrapper_struct!(GidMapping(IdMapping), ParseGidMappingError);

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum ParseFuseOverlayOptionError {
    MaxIdleThreads(ParseMaxIdleThreadsError),
    MaxThreads(ParseMaxThreadsError),
    SquashToUid(ParseSquashToUidError),
    SquashToGid(ParseSquashToGidError),
    UidMapping(ParseUidMappingError),
    GidMapping(ParseGidMappingError),
    UnknownOption(String),
}

impl Display for ParseFuseOverlayOptionError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::MaxIdleThreads(e) => write!(f, "{e}"),
            Self::MaxThreads(e) => write!(f, "{e}"),
            Self::SquashToUid(e) => write!(f, "{e}"),
            Self::SquashToGid(e) => write!(f, "{e}"),
            Self::UidMapping(e) => write!(f, "{e}"),
            Self::GidMapping(e) => write!(f, "{e}"),
            Self::UnknownOption(opt) => write!(f, "unknown option: {opt}"),
        }
    }
}

impl Error for ParseFuseOverlayOptionError {}

impl_from_variants!(
    ParseFuseOverlayOptionError,
    MaxIdleThreads(ParseMaxIdleThreadsError),
    MaxThreads(ParseMaxThreadsError),
    SquashToUid(ParseSquashToUidError),
    SquashToGid(ParseSquashToGidError),
    UidMapping(ParseUidMappingError),
    GidMapping(ParseGidMappingError)
);

#[derive(Clone, Debug, PartialEq, Eq)]
pub enum FuseOverlayOption {
    LowerDir(LowerDirs),
    UpperDir(PathBuf),
    WorkDir(PathBuf),
    CloneFd,
    MaxIdleThreads(MaxIdleThreads),
    MaxThreads(MaxThreads),
    AllowOther,
    AllowRoot,
    SquashToRoot,
    SquashToUid(SquashToUid),
    SquashToGid(SquashToGid),
    StaticNLink,
    NoAcl,
    UidMapping(UidMapping),
    GidMapping(GidMapping),
}

impl_from_variants!(
    FuseOverlayOption,
    LowerDir(LowerDirs),
    MaxIdleThreads(MaxIdleThreads),
    MaxThreads(MaxThreads),
    SquashToUid(SquashToUid),
    SquashToGid(SquashToGid),
    UidMapping(UidMapping),
    GidMapping(GidMapping)
);

impl FromStr for FuseOverlayOption {
    type Err = ParseFuseOverlayOptionError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let s = s.trim().trim_start_matches('-');

        let option = match s.split_once([' ', '=']) {
            Some((option, value)) => (option, Some(value)),
            None => (s, None),
        };

        match option {
            ("lowerdir", Some(args)) => Ok(LowerDirs::from_str(args).unwrap().into()),
            ("upperdir", Some(args)) => Ok(Self::UpperDir(PathBuf::from(args))),
            ("workdir", Some(args)) => Ok(Self::WorkDir(PathBuf::from(args))),
            ("clonefd", None) => Ok(Self::CloneFd),
            ("max_idle_threads", Some(args)) => Ok(MaxIdleThreads::from_str(args)?.into()),
            ("max_threads", Some(args)) => Ok(MaxThreads::from_str(args)?.into()),
            ("allow_other", None) => Ok(Self::AllowOther),
            ("allow_root", None) => Ok(Self::AllowRoot),
            ("squash_to_root", None) => Ok(Self::SquashToRoot),
            ("squash_to_uid", Some(uid)) => Ok(SquashToUid::from_str(uid)?.into()),
            ("squash_to_gid", Some(gid)) => Ok(SquashToGid::from_str(gid)?.into()),
            ("static_nlink", None) => Ok(Self::StaticNLink),
            ("noacl", None) => Ok(Self::NoAcl),
            ("uidmapping", Some(ids)) => Ok(UidMapping::from_str(ids)?.into()),
            ("gidmapping", Some(ids)) => Ok(GidMapping::from_str(ids)?.into()),
            (opt, _) => Err(ParseFuseOverlayOptionError::UnknownOption(opt.to_string())),
        }
    }
}

impl Display for FuseOverlayOption {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "-o ")?;

        match self {
            FuseOverlayOption::LowerDir(lower) => write!(f, "lowerdir={lower}"),
            FuseOverlayOption::UpperDir(path) => write!(f, "upperdir={}", path.to_string_lossy()),
            FuseOverlayOption::WorkDir(path) => write!(f, "workdir={}", path.to_string_lossy()),
            FuseOverlayOption::CloneFd => write!(f, "clone_fd"),
            FuseOverlayOption::MaxIdleThreads(n) => write!(f, "max_idle_threads={n}"),
            FuseOverlayOption::MaxThreads(n) => write!(f, "max_threads={n}"),
            FuseOverlayOption::AllowOther => write!(f, "allow_other"),
            FuseOverlayOption::AllowRoot => write!(f, "allow_root"),
            FuseOverlayOption::SquashToRoot => write!(f, "squash_to_root"),
            FuseOverlayOption::SquashToUid(uid) => write!(f, "squash_to_uid={uid}"),
            FuseOverlayOption::SquashToGid(gid) => write!(f, "squash_to_gid={gid}"),
            FuseOverlayOption::StaticNLink => write!(f, "static_nlink"),
            FuseOverlayOption::NoAcl => write!(f, "noacl"),
            FuseOverlayOption::UidMapping(map) => write!(f, "uidmapping={map}"),
            FuseOverlayOption::GidMapping(map) => write!(f, "gidmapping={map}"),
        }
    }
}

#[derive(Debug)]
pub struct FuseOverlay {
    mount_target: PathBuf,
    options: MountOptions<FuseOverlayOption>,
}

impl FuseOverlay {
    pub fn new(target: impl Into<PathBuf>) -> Self {
        Self {
            mount_target: target.into(),
            options: Default::default(),
        }
    }

    pub fn push_option(&mut self, option: FuseOverlayOption) {
        self.options.push(option);
    }

    pub fn push_options(&mut self, options: impl IntoIterator<Item = FuseOverlayOption>) {
        self.options.extend(options)
    }

    fn find_option<'a, T, F>(&'a self, f: F) -> Option<T>
    where
        F: Fn(&'a FuseOverlayOption) -> Option<T>,
    {
        self.options.iter().filter_map(f).take(1).next()
    }
}

impl Stack for FuseOverlay {
    fn lower_dirs(&self) -> impl Iterator<Item = &std::path::Path> {
        let result = self.find_option(|opt| match opt {
            FuseOverlayOption::LowerDir(dirs) => Some(dirs),
            _ => None,
        });

        match result {
            Some(dirs) => dirs.into_iter(),
            None => DirIter::default(),
        }
    }

    fn upper_dir(&self) -> Option<&std::path::Path> {
        self.find_option(|opt| match opt {
            FuseOverlayOption::UpperDir(dir) => Some(dir.as_path()),
            _ => None,
        })
    }

    fn work_dir(&self) -> Option<&std::path::Path> {
        self.find_option(|opt| match opt {
            FuseOverlayOption::WorkDir(dir) => Some(dir.as_path()),
            _ => None,
        })
    }
}

impl Mountpoint for FuseOverlay {
    fn options(&self) -> impl Iterator<Item = impl Display> {
        self.options.as_iter()
    }
}

impl FileSystem for FuseOverlay {
    fn mount(&mut self) -> std::io::Result<()> {
        Command::new("fuse-overlayfs")
            .arg(self.options.to_string())
            .output()?;

        Ok(())
    }

    fn unmount(&mut self) -> std::io::Result<()> {
        Command::new("fusermount")
            .arg("-u")
            .arg(&self.mount_target)
            .output()?;

        Ok(())
    }
}

impl Display for FuseOverlay {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "fuse-overlayfs")?;

        // INFO: this is to get rid of the extraneous space if
        // self.options is empty
        if self.options.len() > 0 {
            write!(f, " {}", self.options)?;
        }

        write!(f, " {}", self.mount_target.to_string_lossy())
    }
}

#[cfg(test)]
mod tests {
    use std::path::Path;
    use std::str::FromStr;

    use super::*;
    use crate::fs::id_mapping::UntypedIdRange;
    use crate::fs::stackable::Stack;
    use crate::fs::Mountpoint;

    #[test]
    fn lower_dirs() {
        let empty = LowerDirs::default();
        assert_eq!(empty.to_string(), "");

        let one = LowerDirs::from_iter(["/lower"]);
        assert_eq!(one.to_string(), "/lower");

        let two = LowerDirs::from_iter(["/first", "/second"]);
        assert_eq!(two.to_string(), "/first,/second");

        let from_str_none = LowerDirs::from_str("").unwrap();
        assert_eq!(from_str_none, empty);

        let from_str_one = LowerDirs::from_str("/lower").unwrap();
        assert_eq!(from_str_one, one);

        let from_str_two = LowerDirs::from_str("/first,/second").unwrap();
        assert_eq!(from_str_two, two);
    }

    #[test]
    fn max_threads() {
        let default_threads = MaxThreads::default();
        assert_eq!(*default_threads, 10);
        assert_eq!(default_threads.to_string(), "10");

        assert_eq!(MaxThreads::from_str("10").unwrap(), default_threads);
        let err = MaxThreads::from_str("");
        assert!(matches!(err, Err(ParseMaxThreadsError(_))));
        assert_eq!(MaxThreads::from_str("69").unwrap(), MaxThreads(69));
    }

    #[test]
    fn max_idle_threads() {
        let default_threads = MaxIdleThreads::default();
        assert_eq!(*default_threads, None);
        assert_eq!(default_threads.to_string(), "-1");
        assert_eq!(MaxIdleThreads::from_str("-1").unwrap(), default_threads);
        assert_eq!(
            MaxIdleThreads::from_str("69").unwrap(),
            MaxIdleThreads(Some(69))
        );
        let err = MaxIdleThreads::from_str("");
        assert!(matches!(err, Err(ParseMaxIdleThreadsError(_))));
    }

    #[test]
    fn squash_to_id() {
        let squash = SquashToId(1);
        assert_eq!(squash.to_string(), "1");
        assert_eq!(SquashToId::from_str("1").unwrap(), squash);
        let err = SquashToId::from_str("");
        assert!(matches!(err, Err(ParseSquashToIdError(_))));

        let uid = SquashToUid(SquashToId(2));
        let gid = SquashToGid(SquashToId(4));

        assert_eq!(uid.to_string(), "2");
        assert_eq!(gid.to_string(), "4");

        assert_eq!(SquashToUid::from_str("2").unwrap(), uid);
        assert_eq!(SquashToGid::from_str("4").unwrap(), gid);

        let uid_err = SquashToUid::from_str("");
        let gid_err = SquashToGid::from_str("");
        assert!(matches!(uid_err, Err(ParseSquashToUidError(_))));
        assert!(matches!(gid_err, Err(ParseSquashToGidError(_))));
    }

    #[test]
    fn id_mapping() {
        let id_range = IdRange::Untyped(UntypedIdRange(1, 1000, 1));
        let uid = UidMapping(IdMapping::new([id_range.clone()]));
        let gid = GidMapping(IdMapping::new([id_range.clone()]));

        assert_eq!(uid.to_string(), "1:1000:1");
        assert_eq!(gid.to_string(), "1:1000:1");

        assert_eq!(UidMapping::from_str("1:1000:1").unwrap(), uid);
        assert_eq!(GidMapping::from_str("1:1000:1").unwrap(), gid);
    }

    #[test]
    fn command_output() {
        let basic = FuseOverlay::new("/mnt");
        assert_eq!(basic.to_string(), "fuse-overlayfs /mnt");

        let mut with_options = FuseOverlay::new("~/.config");
        with_options.push_options([
            FuseOverlayOption::NoAcl,
            LowerDirs::from_iter(["~/.themes", "~/.colors"]).into(),
        ]);

        assert_eq!(
            with_options.to_string(),
            "fuse-overlayfs -o noacl -o lowerdir=~/.themes,~/.colors ~/.config"
        );
    }

    #[test]
    fn mountpoint_trait_impl() {
        let mut all = FuseOverlay::new("/mnt");

        all.push_options([
            LowerDirs::from_iter(["/doesnt", "/matter"]).into(),
            FuseOverlayOption::UpperDir("/upper".into()),
            FuseOverlayOption::WorkDir("/work".into()),
        ]);

        let opts: Vec<String> = all.options().map(|x| x.to_string()).collect();

        assert_eq!(
            opts,
            [
                "-o lowerdir=/doesnt,/matter",
                "-o upperdir=/upper",
                "-o workdir=/work"
            ]
        );
    }

    #[test]
    fn stack_trait_impl() {
        let mut all = FuseOverlay::new("/mnt");

        assert_eq!(all.lower_dirs().next(), None);
        assert_eq!(all.upper_dir(), None);
        assert_eq!(all.work_dir(), None);

        all.push_option(LowerDirs::from_iter(["/lower"]).into());
        assert_eq!(all.lower_dirs().next(), Some(Path::new("/lower")));

        all.push_option(FuseOverlayOption::UpperDir("/upper".into()));
        assert_eq!(all.upper_dir(), Some(Path::new("/upper")));

        all.push_option(FuseOverlayOption::WorkDir("/work".into()));
        assert_eq!(all.work_dir(), Some(Path::new("/work")));
    }
}
